@@ -1,6 +1,7 @@
 #!/usr/bin/env python3
 
 import collections
+import functools
 import math
 import operator
 import statistics
@@ -81,10 +82,6 @@ def dugast_u(text):
     Slatkine.
 
     """
-    # as suggested by Andreas, we decrease the vocabulary size by 1,
-    # if we only have hapaxes
-    # if text_length == vocabulary_size:
-    #     vocabulary_size -= 1
     if text.text_length == text.vocabulary_size:
         return math.nan
     return (math.log(text.text_length) ** 2) / (math.log(text.text_length) - math.log(text.vocabulary_size))
@@ -101,7 +98,7 @@ def tuldava_ln(text):
     return (1 - (text.vocabulary_size ** 2)) / ((text.vocabulary_size ** 2) * math.log(text.text_length))
 
 
-def brunet_w(text, *, a=-0.172):
+def brunet_w(text, *, a=0.172):
     """Brunet (1978)
 
     Brunet, Étienne (1978). Le vocabulaire de Jean Giraudoux.
@@ -176,8 +173,6 @@ def honore_h(text):
     Bulletin 7 (2), pp. 172–177.
 
     """
-    # Similar to dugast_u, we decrease the number of hapax legomena by
-    # 1, if we only have hapaxes
     hapaxes = text.frequency_spectrum.get(1, 0)
     try:
         return 100 * (math.log(text.text_length) / (1 - (hapaxes / text.vocabulary_size)))
@@ -203,6 +198,19 @@ def evenness(text):
     return entropy(text) / math.log2(text.vocabulary_size)
 
 
+def jarvis_evenness(text):
+    """Jarvis (2013: 29) suggests “to use the standard deviation of the
+    number of tokens per type in a text” as a measure of evenness.
+
+    Jarvis, Scott (2013). “Defining and measuring lexical diversity”.
+    In: Jarvis, Scott and Daller, Michael, Vocabulary Knowledge. Human
+    ratings and automated measures. John Benjamins. 13--44.
+    https://doi.org/10.1075/sibil.47.03ch1.
+
+    """
+    return statistics.stdev(text.frequency_list.values())
+
+
 def yule_k(text):
     """Yule (1944)
 
@@ -320,7 +328,7 @@ def _gries_dp(text, n_parts):
     part_size = text.text_length // n_parts
     s_percentage_of_part = 1 / n_parts
     frequencies = np.zeros((text.vocabulary_size, n_parts))
-    word_idx = {t: i for i, t in enumerate(sorted(set(text.tokens)))}
+    word_idx = {t: i for i, t in enumerate(sorted(text.frequency_list.keys()))}
     for i in range(n_parts):
         part = text.tokens[i * part_size:(i * part_size) + part_size]
         for token, freq in collections.Counter(part).items():
@@ -378,20 +386,83 @@ def kl_divergence(text, n_parts):
     """
     part_size = text.text_length // n_parts
     frequencies = np.zeros((text.vocabulary_size, n_parts))
-    word_idx = {t: i for i, t in enumerate(sorted(set(text.tokens)))}
+    word_idx = {t: i for i, t in enumerate(sorted(text.frequency_list.keys()))}
     for i in range(n_parts):
         part = text.tokens[i * part_size:(i * part_size) + part_size]
         for token, freq in collections.Counter(part).items():
             frequencies[word_idx[token], i] = freq
     # kld_scores = {token: sum([0 if v[token] == 0 else (v[token] / frequency) * math.log2((v[token] / frequency) * n_parts) for v in v_frequency_corpus_part]) for token, frequency in f_overall_frequency.items()}
     relative_frequencies = np.divide(frequencies, np.sum(frequencies, axis=1).reshape(-1, 1))
-    with np.errstate(divide='ignore'):
-        log = np.log2(np.multiply(relative_frequencies, n_parts))
-    log[np.isneginf(log)] = 0
+    log = np.log2(np.where(frequencies != 0, np.multiply(relative_frequencies, n_parts), np.ones_like(relative_frequencies)))
     kld_scores = np.sum(np.multiply(relative_frequencies, log), axis=1)
     return np.mean(kld_scores)
 
 
+def _get_distances_for_gini(text):
+    distances = {t: [] for t in text.frequency_list.keys()}
+    first, last = {}, {}
+    for i, token in enumerate(text.tokens):
+        if token in first:
+            dist = i - last[token]
+            distances[token].append(dist)
+        else:
+            first[token] = i
+        last[token] = i
+    for token in distances.keys():
+        dist = first[token] + text.text_length - last[token]
+        distances[token].append(dist)
+    return {t: np.array(dists) for t, dists in distances.items()}
+
+
+def gini_based_dispersion(text, exclude_hapaxes=False):
+    """This measure indicates how evenly the distances between tokens of
+    the same type are distributed. Gini-based dispersion for a single
+    type is computed as 1 - (Gini / Gini_max), where Gini is the Gini
+    coefficient (https://en.wikipedia.org/wiki/Gini_coefficient) of
+    the distances and Gini_max = (N - f) * (f - 1) / (f * N) is the
+    maximum value for a type with frequency f in a text of length N.
+    This function returns the arithmetic mean of the Gini-based
+    dispersion scores for all types in the text. Returns math.nan if
+    exclude_hapaxes=True and the text only consists of hapax legomena.
+
+    """
+    distances = _get_distances_for_gini(text)
+    ginis = {t: scipy.spatial.distance.squareform(scipy.spatial.distance.pdist(dists.reshape(-1, 1))).mean() / (2 * dists.mean()) for t, dists in distances.items()}
+    gini_max = {t: (text.text_length - f) * (f - 1) / (f * text.text_length) for t, f in text.frequency_list.items()}
+    dispersions = {t: 1 if g == 0 else 1 - (g / gini_max[t]) for t, g in ginis.items()}
+    if exclude_hapaxes:
+        dispersions = {t: d for t, d in dispersions.items() if text.frequency_list[t] > 1}
+    try:
+        return statistics.mean(dispersions.values())
+    except statistics.StatisticsError:
+        return math.nan
+
+
+def evenness_based_dispersion(text, exclude_hapaxes=False):
+    """This measure indicates how evenly the distances between tokens of
+    the same type are distributed based on evenness (also known as
+    normalized entropy or efficiency). Evenness-based dispersion for a
+    single type is computed as (Evenness - Evenness_min) / (1 -
+    Evenness_min), Evenness_min = ((1 - f) * (1 / N) * log2(1 / N) -
+    ((N - f + 1) / N)) * log2((N - f + 1) / N) / log2(f) is the
+    minimum value for a type with frequency f in a text of length N.
+    This function returns the arithmetic mean of the Evenness-based
+    dispersion scores for all types in the text. Returns math.nan if
+    exclude_hapaxes=True and the text only consists of hapax legomena.
+
+    """
+    distances = _get_distances_for_gini(text)
+    evenness_scores = {t: (-1 * np.multiply(np.divide(dists, text.text_length), np.log2(np.divide(dists, text.text_length))).sum()) / math.log2(text.frequency_list[t]) if text.frequency_list[t] > 1 else 0 for t, dists in distances.items()}
+    evenness_min = {t: ((1 - f) * (1 / text.text_length) * math.log2(1 / text.text_length) - ((text.text_length - f + 1) / text.text_length) * math.log2((text.text_length - f + 1) / text.text_length)) / math.log2(f) if f > 1 else 0 for t, f in text.frequency_list.items()}
+    dispersions = {t: (e - evenness_min[t]) / (1 - evenness_min[t]) for t, e in evenness_scores.items()}
+    if exclude_hapaxes:
+        dispersions = {t: d for t, d in dispersions.items() if text.frequency_list[t] > 1}
+    try:
+        return statistics.mean(dispersions.values())
+    except statistics.StatisticsError:
+        return math.nan
+
+
 # ---------------------------------- #
 # PARAMETERS OF PROBABILISTIC MODELS #
 # ---------------------------------- #
@@ -454,29 +525,4 @@ def sttr(tokens, window_size=1000, strategy="spread"):
     http://purl.org/stefan.evert/PUB/EvertWankerlNoeth2017.pdf
 
     """
-    return bootstrap(type_token_ratio, tokens, window_size, strategy)
-
-
-# ------------- #
-# BOOTSTRAPPING #
-# ------------- #
-
-def bootstrap(measure, tokens, window_size, strategy="spread", **kwargs):
-    """Calculate bootstrap for surface-based measures as explained in
-    Evert et al. (2017).
-
-    kwargs are passed to measure
-
-    Evert, Stefan, Sebastian Wankerl, Elmar Nöth (2017). Reliable
-    measures of syntactic and lexical complexity: The case of Iris
-    Murdoch. In: Proceedings of the Corpus Linguistics 2017
-    Conference, Birmingham, UK.
-    http://purl.org/stefan.evert/PUB/EvertWankerlNoeth2017.pdf
-
-    """
-    results = []
-    for window in windows.disjoint_windows(tokens, window_size, strategy):
-        results.append(measure(window, **kwargs))
-    if len(results) == 1:
-        return results[0], 0, results
-    return statistics.mean(results), misc.confidence_interval(results), results
+    return misc.bootstrap(type_token_ratio, tokens, window_size, strategy)
